/* expand-core.c - elementary functions for the expand and unexpand utilities
   Copyright (C) 1989-2015 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include <config.h>

#include <stdio.h>
#include <sys/types.h>
#include <mbfile.h>

#include "system.h"
#include "error.h"
#include "fadvise.h"
#include "quote.h"
#include "xstrndup.h"

#include "expand-core.h"

extern inline int
set_utf_locale (void)
{
      /*try using some predefined locale */
      const char* predef_locales[] = {"C.UTF8","en_US.UTF8","en_GB.UTF8"};

      const int predef_locales_count=3;
      for (int i=0;i<predef_locales_count;i++)
        {
          if (setlocale(LC_ALL,predef_locales[i])!=NULL)
          {
            break;
          }
          else if (i==predef_locales_count-1)
          {
            return 1;
            error (EXIT_FAILURE, errno, _("cannot set UTF-8 locale"));
          }
        }
        return 0;
}

extern bool
check_utf_locale(void)
{
  char* locale = setlocale (LC_CTYPE , NULL);
  if (locale == NULL) 
  {
    return false;
  }
  else if (strcasestr(locale, "utf8") == NULL && strcasestr(locale, "utf-8") == NULL)
  { 
    return false;
  }
  return true;
}

extern bool
check_bom(FILE* fp, mb_file_t *mbf)
{
  int c;


  c=fgetc(fp);

  /*test BOM header of the first file */
  mbf->bufcount=0;
  if (c == 0xEF)
  {
    c=fgetc(fp);
  }
  else
  {
    if (c != EOF)
    {
      ungetc(c,fp);
    }
    return false;
  }

  if (c == 0xBB)
  {
    c=fgetc(fp);
  }
  else
  {
    if ( c!= EOF )
    {
      mbf->buf[0]=(unsigned char) 0xEF;
      mbf->bufcount=1;
      ungetc(c,fp);
      return false;
    }
    else
    {
      ungetc(0xEF,fp);
      return false;
    }
  }
  if (c == 0xBF)
  {
    mbf->bufcount=0;
    return true;
  }
  else
  {
    if (c != EOF)
    {
      mbf->buf[0]=(unsigned char) 0xEF;
      mbf->buf[1]=(unsigned char) 0xBB;
      mbf->bufcount=2;
      ungetc(c,fp);
      return false;
    }
    else
    {
      mbf->buf[0]=(unsigned char) 0xEF;
      mbf->bufcount=1;
      ungetc(0xBB,fp);
      return false;
    }
  }
  return false;
}

extern inline void
print_bom(void)
{
  putc (0xEF, stdout);
  putc (0xBB, stdout);
  putc (0xBF, stdout);
}

/* Add the comma or blank separated list of tab stops STOPS
   to the list of tab stops.  */

extern void
parse_tab_stops (char const *stops, void (*add_tab_stop)(uintmax_t))
{
  bool have_tabval = false;
  uintmax_t tabval IF_LINT ( = 0);
  char const *num_start IF_LINT ( = NULL);
  bool ok = true;

  for (; *stops; stops++)
    {
      if (*stops == ',' || isblank (to_uchar (*stops)))
        {
          if (have_tabval)
            add_tab_stop (tabval);
          have_tabval = false;
        }
      else if (ISDIGIT (*stops))
        {
          if (!have_tabval)
            {
              tabval = 0;
              have_tabval = true;
              num_start = stops;
            }

          /* Detect overflow.  */
          if (!DECIMAL_DIGIT_ACCUMULATE (tabval, *stops - '0', uintmax_t))
            {
              size_t len = strspn (num_start, "0123456789");
              char *bad_num = xstrndup (num_start, len);
              error (0, 0, _("tab stop is too large %s"), quote (bad_num));
              free (bad_num);
              ok = false;
              stops = num_start + len - 1;
            }
        }
      else
        {
          error (0, 0, _("tab size contains invalid character(s): %s"),
                 quote (stops));
          ok = false;
          break;
        }
    }

  if (!ok)
    exit (EXIT_FAILURE);

  if (have_tabval)
    add_tab_stop (tabval);
}

/* Check that the list of tab stops TABS, with ENTRIES entries,
   contains only nonzero, ascending values.  */

extern void
validate_tab_stops (uintmax_t const *tabs, size_t entries)
{
  uintmax_t prev_tab = 0;
  size_t i;

  for (i = 0; i < entries; i++)
    {
      if (tabs[i] == 0)
        error (EXIT_FAILURE, 0, _("tab size cannot be 0"));
      if (tabs[i] <= prev_tab)
        error (EXIT_FAILURE, 0, _("tab sizes must be ascending"));
      prev_tab = tabs[i];
    }
}

/* Close the old stream pointer FP if it is non-NULL,
   and return a new one opened to read the next input file.
   Open a filename of '-' as the standard input.
   Return NULL if there are no more input files.  */

extern FILE *
next_file (FILE *fp)
{
  static char *prev_file;
  char *file;

  if (fp)
    {
      if (ferror (fp))
        {
          error (0, errno, "%s", prev_file);
          exit_status = EXIT_FAILURE;
        }
      if (STREQ (prev_file, "-"))
        clearerr (fp);		/* Also clear EOF.  */
      else if (fclose (fp) != 0)
        {
          error (0, errno, "%s", prev_file);
          exit_status = EXIT_FAILURE;
        }
    }

  while ((file = *file_list++) != NULL)
    {
      if (STREQ (file, "-"))
        {
          have_read_stdin = true;
          fp = stdin;
        }
      else
        fp = fopen (file, "r");
      if (fp)
        {
          prev_file = file;
          fadvise (fp, FADVISE_SEQUENTIAL);
          return fp;
        }
      error (0, errno, "%s", file);
      exit_status = EXIT_FAILURE;
    }
  return NULL;
}
